--Functions by NeZvers

local TILE_SIZE  = 16				--default size
local TILE_ROUND = -TILE_SIZE + 1	--in case of bitwise calculations
local VIEW_SCALE = 4				--hardcoded for prototyping
local JUMP_BUFFER = 5				--buffer after player loses ground

--COLISION TILES
local solid1 = 1					--solid block
local solid2 = 2					--slope 45 righ
local solid3 = 3					--slope 45 left
local solid4 = 4					--slope 22.5 right 1/2
local solid5 = 5					--slope 22.5 right 2/2
local solid6 = 6					--slope 22.5 left 1/2
local solid7 = 7					--slope 22.5 left 2/2

--will be used as bitmasks to check buttons
local up     = 1
local down   = 2
local left   = 4
local right  = 8
local jump   = 16
local action = 32
local start  = 64

function debug_on(peakTrue)
	timer.delay(1, false, function()
		profiler.enable_ui(true)
		profiler.set_ui_mode(profiler.MODE_RECORD)
		if peakTrue then
			profiler.set_ui_mode(profiler.MODE_SHOW_PEAK_FRAME) -- comment this line to not show peak only
		end
	end)
end

function clamp(v, min, max)
	if v < min then v = min 
	elseif v > max then v = max end
	return v
end

function approach(start, ending, ammount)
	local result = nil
	if start > ending then 
		if start - ammount < ending then result = ending
		else result = start - ammount end
	elseif start < ending then
		if start + ammount > ending then result = ending
		else result = start + ammount end
	else result = ending end
	return result
end

function approach_alt(start, ending, ammount)
	local result = nil
	if start < ending then
		result = math.min(start + ammount, ending)
	else result = math.max(start - ammount, ending) end
	return result
end

function sign(v)
	if v > 0 then return 1
	elseif v < 0 then return -1
	else return 0 end
end

function div(a,b)
	return (a - a % b) / b
end

function round(v)
	return v + 0.5 - (v + 0.5) % 1
end

local ceil	= math.ceil																--Save floor function for simplicity
local floor	= math.floor															--save ceil function for simplicity

function set_hitbox(inst, hitbox_r, hitbox_l, hitbox_t, hitbox_b)
	inst.hitbox_l  = hitbox_l
	inst.hitbox_r  = hitbox_r
	inst.hitbox_t  = hitbox_t
	inst.hitbox_b  = hitbox_b
	inst.hitbox_hc = math.ceil((inst.hitbox_r + inst.hitbox_l)/2)
	inst.hitbox_vc = math.ceil((inst.hitbox_t + inst.hitbox_b)/2)
end

function init_physics(inst, id, url, solidMap, solidLayer, tile_size, run_maxspeed, jump_speed, gravity)
	TILE_SIZE         = tile_size
	TILE_ROUND		  = -TILE_SIZE + 1
	inst.MY_ID		  = id
	inst.MY_URL		  = url
	inst.MY_POS		  = go.get_position(url)
	inst.START_POS	  = go.get_position(url)
	inst.SOLIDMAP     = solidMap
	inst.SOLIDLAYER   = solidLayer
	inst.GRAVITY      = -gravity

	inst.START_JUMP   = jump_speed
	inst.RELEASE_JUMP = ceil(jump_speed / 3)
	inst.ACC          = run_maxspeed / 10		--Accelerate
	inst.ACCW         = run_maxspeed / 30		--Walk accelerate
	inst.MAX          = run_maxspeed			--max run speed
	inst.MAXW         = run_maxspeed  * 25 / 64	--Max walk speed
	inst.DCC          = run_maxspeed * 9 / 64	--deaccelerate when no input
	--inst.AIR          = run_maxspeed * 2 / 64
	--inst.AIRW         = run_maxspeed * 1 / 128	--air walk speed
	--inst.AIRS         = run_maxspeed * 2 / 64 	--air stopping speed
	--inst.DRAG         = run_maxspeed * 9 / 64	--deaccelerate when no input
	inst.MAX_DOWN     = -jump_speed				--max fall speed
	inst.MAX_UP       = jump_speed				--max up speed just in case
	inst.WALLSPEED	  = -ceil(jump_speed/5)

	inst.spd          = vmath.vector3(0,0,0)
	inst.fraction     = vmath.vector3(0,0,0)

	inst.buttons	  = 0
	inst.xinput		  = 0								--for platformer
	--inst.dir_input	  = vmath.vector3(0, 0, 0)		--for top-down
	inst.jmp_buf_tmr  = 0
	inst.lastWall     = 0
	inst.hurtTimer    = 0
	inst.dashTimer    = 0
	
	--STATES
	inst.grounded	  = false
	inst.landed		  = false	--not included yet
	inst.jumping	  = false
	inst.doubleJump   = false	--not included yet
	inst.dashing	  = false	--not included yet
	inst.wallsliding  = false
	inst.onLedge	  = false	--not included yet
	inst.climbing	  = false	--not included yet
	inst.hurt		  = false	--not included yet

	--HITBOX
	inst.hitbox_l	  = 0
	inst.hitbox_r	  = 0
	inst.hitbox_t	  = 0
	inst.hitbox_b	  = 0
	inst.hitbox_hc    = math.ceil((inst.hitbox_r + inst.hitbox_l)/2)
	inst.hitbox_vc    = math.ceil((inst.hitbox_t + inst.hitbox_b)/2)
end

function button_up(inst)     inst.buttons = bit.bor(inst.buttons, up)     end
function button_down(inst)   inst.buttons = bit.bor(inst.buttons, down)   end
function button_left(inst)   inst.buttons = bit.bor(inst.buttons, left)   end
function button_right(inst)  inst.buttons = bit.bor(inst.buttons, right)  end
function button_jump(inst)   inst.buttons = bit.bor(inst.buttons, jump)   end
function button_action(inst) inst.buttons = bit.bor(inst.buttons, action) end
function button_start(inst)  inst.buttons = bit.bor(inst.buttons, start)  end

function get_xinput(inst)
	local hin = 0
	if (bit.band(inst.buttons, right) > 0) then										--bitmasking for right button
		hin = hin + 1
	end
	if (bit.band(inst.buttons, left) > 0) then										--bitmasking for left button
		hin = hin - 1
	end
	inst.xinput = hin
end
--Get Tile ID from tilesource
function tile_id(inst, x, y)
	x = ceil(x)
	y = ceil(y)
	return tilemap.get_tile(inst.SOLIDMAP, inst.SOLIDLAYER, math.ceil(x/TILE_SIZE), math.ceil(y/TILE_SIZE))
end

--Get y one pixel above the tile
function tile_height(inst, tile_id, x, y)
	x = ceil(x)
	y = ceil(y)
	if tile_id == solid1 then																							--Block tile
		return ceil(y/TILE_SIZE) * TILE_SIZE + 1
	elseif tile_id == solid2 then																						--45 /
		return floor((y-1)/TILE_SIZE) * TILE_SIZE +1 +((x-1) % TILE_SIZE)
	elseif tile_id == solid3 then																						--45 \
		return floor((y-1)/TILE_SIZE) * TILE_SIZE +TILE_SIZE -((x-1) % TILE_SIZE)
	elseif tile_id == solid4 then																						--22.5 / low
		return floor((y-1)/TILE_SIZE) * TILE_SIZE + floor((x -1 - div(x-1, TILE_SIZE)*TILE_SIZE)/2) +1
	elseif tile_id == solid5 then																						--22.5 / high
		return floor((y-1)/TILE_SIZE) * TILE_SIZE + floor((x-1 - div(x-1, TILE_SIZE)*TILE_SIZE)/2) +floor(TILE_SIZE/2) +1
	elseif tile_id == solid6 then																						--22.5 \ low
		return floor((y-1)/TILE_SIZE) * TILE_SIZE +floor(TILE_SIZE/2) - floor((x -1 - div(x-1, TILE_SIZE)*TILE_SIZE)/2)
	elseif tile_id == solid7 then																						--22.5 \ high
		return floor((y-1)/TILE_SIZE) * TILE_SIZE +TILE_SIZE - floor((x -1 - div(x-1, TILE_SIZE)*TILE_SIZE)/2)
	elseif tile_id == 0  then																							--Empty
		return y
	elseif tile_id == nil  then																							--Out of map
		return y
	end
end

function ground_check_blocks(inst)
	--local x = inst.MY_POS.x
	--local y = inst.MY_POS.y
	local x = ceil(inst.MY_POS.x)
	local y = ceil(inst.MY_POS.y)
	if inst.spd.y <= 0 then															--Bypass in case going up
		local L = tile_id(inst, x+inst.hitbox_l+1,	y+inst.hitbox_b)
		local M = tile_id(inst, x+inst.hitbox_hc,	y+inst.hitbox_b)
		local R = tile_id(inst, x+inst.hitbox_r,	y+inst.hitbox_b)
		if L>0 or M>0 or R>0 then
			inst.grounded = true
		else
			inst.grounded = false
		end
	else
		inst.grounded = false
	end
end

function ground_check_slopes(inst)
	--local x = inst.MY_POS.x
	--local y = inst.MY_POS.y
	local x = ceil(inst.MY_POS.x)
	local y = ceil(inst.MY_POS.y)
	inst.grounded = false															--Default to false
	if inst.spd.y <= 0 then															--Bypass if going up
		local M = tile_id(inst, x+inst.hitbox_hc, y+inst.hitbox_b)					--Get tile id
		if M~=nil and M>0 then														--Center bottom is inside solid tile
			local h = tile_height(inst, M, x+inst.hitbox_hc, y+inst.hitbox_b)
			if y+inst.hitbox_b < h then												--If feet is on or below tile height
				inst.grounded = true
			end
		end
		if inst.grounded == false then												--If middle isn't on ground
			local L = tile_id(inst, x+inst.hitbox_l+1, y+inst.hitbox_b)
			local R = tile_id(inst, x+inst.hitbox_r, y+inst.hitbox_b)
			if L~=nil and L~=0 then
				local h = tile_height(inst, L, x+inst.hitbox_l+1, y+inst.hitbox_b)
				if y+inst.hitbox_b < h then
					inst.grounded = true
				end
			end
			if not inst.grounded and L~=nil and R~=0 then
				local h = tile_height(inst, R, x+inst.hitbox_r, y+inst.hitbox_b)
				if y+inst.hitbox_b < h then
					inst.grounded = true
				end
			end
		end
	end
end

function can_wallslide(inst, h_dir)													--Check if can wallslide
	if not inst.wallsliding then													--Not attached
		if h_dir==0 then															--Not pressing any direction
			return false
		else
			if inst.lastWall == inst.xinput then									--Have latched to wall this direction (simple solution to disable same wall)
				return false
			else
				local side = nil
				if h_dir>0 then
					side = inst.hitbox_r+1
				elseif h_dir<0 then
					side = inst.hitbox_l
				end
				local x = inst.MY_POS.x
				local T = tile_id(inst, inst.MY_POS.x+side, inst.MY_POS.y+inst.hitbox_t)
				if T==solid1 then													--If tile is solid block
					if h_dir>0 then													--Save this direction for same wall latch disabling
						inst.lastWall = 1
					elseif h_dir<0 then
						inst.lastWall = -1
					end
					return true
				else
					return false
				end
			end
		end
	else																			--Already latched to the wall
		local side = nil
		if inst.lastWall>0 then
			side = inst.hitbox_r+1
		elseif inst.lastWall<0 then
			side = inst.hitbox_l
		end
		local x = inst.MY_POS.x
		local T = tile_id(inst, inst.MY_POS.x+side, inst.MY_POS.y+inst.hitbox_t)	--Check tile ID
		if T==solid1 then																--Tile is solid block
			return true
		else																		--Tile is not solid block
			return false
		end
	end
	
end

function h_move(inst)
	if inst.wallsliding then														--Wallslide
		--
	else
		local hin = inst.xinput														--xinput set in get_xinput()
		local hsp = inst.spd.x
		if hin ~= 0 then															--Move
			hsp = hsp + hin * inst.ACC
			hsp = clamp(hsp, -inst.MAX, inst.MAX)
		else																		--deaccelerate
			hsp = approach(hsp, 0, inst.DCC)										--(value, goal, ammount)
		end
		inst.spd.x = hsp
	end
end

function h_collide(inst)
	local hsp  = inst.spd.x
	local frac = inst.fraction.x
	
	--frac = hsp - math.floor(math.abs(hsp)) * sign(hsp)								--Save new fraction ammount
	hsp  = hsp-- - frac																--Leave remainder in integer value
	if hsp ~= 0 then																--Decouple code if not needed
		local x = inst.MY_POS.x
		local y = inst.MY_POS.y
		local side = nil
		if hsp > 0 then
			side = inst.hitbox_r
		else
			side = inst.hitbox_l
		end
		local T = tile_id(inst, x+side+hsp,		y+inst.hitbox_t)					--check top corner
		local M = tile_id(inst, x+side+hsp,		y+inst.hitbox_vc)					--check midle of the side
		local B = tile_id(inst, x+side+hsp,		y+inst.hitbox_b+1)					--check bottom corner

		if B==solid1 or M==solid1 or T==solid1 then
			if hsp > 0 then															--Moving to the right
				x = math.ceil((x+side+hsp)/TILE_SIZE) * TILE_SIZE -TILE_SIZE -side	--Snap to the tiles left side
			else																	-- Moving to the left or idle
				x = math.ceil((x+side+hsp)/TILE_SIZE) * TILE_SIZE -side				--Snap to the tiles right side
			end
			hsp  = 0 --reset speed after collision
			frac = 0
		end

		inst.MY_POS.x = x + hsp 													--Update coordinates variable
		inst.spd.x = hsp --+ frac														--Update speed variable
		--inst.fraction.x = frac														--Save fraction value
	end
end

function v_move_platformer(inst)
	local vsp = inst.spd.y
	if inst.grounded then												--On ground
		inst.doublejump = false											--If use doublejump it's disabled on ground
		inst.lastWall = 0												--Reset Last wall (for allowing walljump from same wall once in a row)
		inst.jmp_buf_tmr = 0											--Reset jump buffer
		inst.lastWall = 0												--Reset last lastched wall direction
		if inst.jumping and bit.band(inst.buttons, jump)==0 then		--Release jump button
			inst.jumping = false
		end
		if inst.wallsliding then										--disable wallsliding
			inst.wallsliding = false
		end	

		if bit.band(inst.buttons, jump)==jump and not inst.jumping and (bit.band(inst.buttons, down)==0) then --New jump executed
			vsp = inst.START_JUMP
			inst.grounded = false
			inst.jumping = true
		end
	else 																--Not on the ground
		vsp = vsp + inst.GRAVITY										--Apply gravity
		if vsp>0 then													--Going up
			if inst.jmp_buf_tmr~=JUMP_BUFFER then						--Disable jump buffer
				inst.jmp_buf_tmr = JUMP_BUFFER
			end
		else															--Going down
			if can_wallslide(inst, inst.xinput) then					--Check wallsliding
				inst.wallsliding = true
				--inst.jumping = false
			elseif inst.wallsliding then
				inst.wallsliding = false
			end
			if inst.wallsliding then
				if bit.band(inst.buttons, jump)==jump and not inst.jumping then --New jump executed
					vsp = inst.START_JUMP
					inst.jumping = true
					inst.wallsliding = false
					if inst.xinput == -inst.lastWall then						--Presing away from wall
						inst.spd.x = inst.xinput * inst.MAX
					end
				end
				if vsp < inst.WALLSPEED then vsp = inst.WALLSPEED end	--Limit wallslide speed
			else
				if vsp < inst.MAX_DOWN then vsp = inst.MAX_DOWN end		--Limit fall speed
			end
			if inst.jmp_buf_tmr < JUMP_BUFFER then						--Count jump buffer timer
				inst.jmp_buf_tmr = inst.jmp_buf_tmr +1
			end
		end
		
		if bit.band(inst.buttons, jump)~=jump then						--Released jump button
			if vsp > inst.RELEASE_JUMP then								--Cut down jump speed
				vsp = inst.RELEASE_JUMP
			end
			inst.jumping = false
		else															--Holding jump button
			if not inst.jumping and inst.jmp_buf_tmr < JUMP_BUFFER then
				vsp = inst.START_JUMP
				inst.jumping = true
			elseif not inst.jumping and inst.doublejump then			--If released jump and allowed doublejump
				vsp = inst.JUMP_START
				inst.jumping = true
				inst.doublejump = false
			end
		end
	end
	inst.spd.y = vsp										--Save vertical speed
end

function v_move_topdown(inst)														--Vertical top down movement
	--Get direction
	local vin = 0
	if (bit.band(inst.buttons, up) > 0) then										--bitmasking for up button
		vin = vin + 1
	end
	if (bit.band(inst.buttons, down) > 0) then										--bitmasking for down button
		vin = vin - 1
	end

	local vsp = inst.spd.y
	if vin ~= 0 then						--move
		vsp = vsp + vin * inst.ACC
		vsp = clamp(vsp, -inst.MAX, inst.MAX)
	else																			--deaccelerate
		vsp = approach(vsp, 0, inst.DCC)											--(value, goal, ammount)
	end
	inst.spd.y = vsp
end

function v_collide_blocks(inst)														--Simple block tile collision
	local vsp  = inst.spd.y
	local frac = inst.fraction.y

	frac = vsp - math.floor(math.abs(vsp)) * sign(vsp)								--Save new fraction ammount
	vsp  = vsp - frac																--Leave remainder in integer value
	--vsp = math.floor(math.abs(vsp)) * sign(vsp)
	if vsp ~= 0 then																--Decouple code if not needed
		local x = inst.MY_POS.x
		local y = inst.MY_POS.y
		local side = nil
		if vsp > 0 then
			side = inst.hitbox_t
		else
			side = inst.hitbox_b
		end
		local R = tile_id(inst, x+inst.hitbox_r,		y+side+vsp)					--check top corner
		local M = tile_id(inst, x+inst.hitbox_hc,		y+side+vsp)					--check midle of the side
		local L = tile_id(inst, x+inst.hitbox_l+1,		y+side+vsp)					--check bottom corner

		if R==solid1 or M==solid1 or L==solid1 then
			if vsp > 0 then															--Moving to the right
				y = math.ceil((y+side+vsp)/TILE_SIZE) * TILE_SIZE -TILE_SIZE -side	--Snap to the tiles left side
			else																	-- Moving to the left or idle
				y = math.ceil((y+side+vsp)/TILE_SIZE) * TILE_SIZE -side				--Snap to the tiles right side
			end
			vsp  = 0 --reset speed after collision
			frac = 0
		end

		inst.MY_POS.y = y + vsp --Update coordinates variable
		inst.spd.y = vsp + frac	--Update speed variable
		inst.fraction.y = frac	--Save fraction value
	end
end

function v_collide_slopes(inst)														--Simple block tile collision
	local x = inst.MY_POS.x
	local y = inst.MY_POS.y
	local vsp  = inst.spd.y
	local frac = inst.fraction.y

	--frac = vsp - math.floor(math.abs(vsp)) * sign(vsp)								--Save new fraction ammount
	vsp  = vsp-- - frac																--Leave remainder in integer value
	if vsp ~= 0 then																--Decouple code if not moving
		if vsp > 0 then																	--Going up
			local side = inst.hitbox_t
			local R = tile_id(inst, x+inst.hitbox_r,		y+side+vsp)					--check top corner
			local M = tile_id(inst, x+inst.hitbox_hc,		y+side+vsp)					--check midle of the side
			local L = tile_id(inst, x+inst.hitbox_l+1,		y+side+vsp)					--check bottom corner

			if R==solid1 or M==solid1 or L==solid1 then
				y = math.ceil((y+side+vsp)/TILE_SIZE) * TILE_SIZE -TILE_SIZE -side	--Snap to the tiles left side
				vsp  = 0 --reset speed after collision
				frac = 0
			end
		else																			--Going down
			local side = inst.hitbox_b+1
			local C = tile_id(inst, x+inst.hitbox_hc, y+side+vsp)
			if C==0 or C==nil then
				local R = tile_id(inst, x+inst.hitbox_r,		y+side+vsp)					--check top corner
				local M = tile_id(inst, x+inst.hitbox_hc,		y+side+vsp)					--check midle of the side
				local L = tile_id(inst, x+inst.hitbox_l+1,		y+side+vsp)					--check bottom corner

				if R==solid1 or M==solid1 or L==solid1 then
					y = math.ceil((y+side+vsp)/TILE_SIZE) * TILE_SIZE -side	+1			--Snap to the tiles right side
					vsp  = 0 --reset speed after collision
					frac = 0
				end
			elseif C==solid1 then
				y = ceil((y+side+vsp)/TILE_SIZE) * TILE_SIZE -side	+1			--Snap to the tiles right side
				vsp  = 0 --reset speed after collision
				frac = 0
			elseif C~=solid1 then
				local h = tile_height(inst, C, x+inst.hitbox_hc, y+inst.hitbox_b+1+vsp)
				if (y+inst.hitbox_b+1+vsp)<=h then
					y = h-(inst.hitbox_b+1)
					vsp  = 0
					frac = 0
				end
			end
		end
	end
	
	if inst.grounded then
		local yy = 1
		local M = tile_id(inst, x+inst.hitbox_hc, y+inst.hitbox_b+yy)
		if M==0 or M==nil then
			yy = 0
			M = tile_id(inst, x+inst.hitbox_hc, y+inst.hitbox_b+yy)
		end
		if M==0 or M==nil then
			yy = TILE_SIZE
			M = tile_id(inst, x+inst.hitbox_hc, y+inst.hitbox_b+yy)
		end
		if M~=0 and M~=nil then
			local h = tile_height(inst, M, x+inst.hitbox_hc, y+inst.hitbox_b+yy)
			y = h -(inst.hitbox_b+1)
			vsp = 0
			frac = 0
		end
	end
	inst.MY_POS.y = y + vsp --Update position variable
	inst.spd.y = vsp-- + frac	--Update speed variable
	--inst.fraction.y = frac	--Save fraction value
end

function physics_update_topdown(inst)
	get_dir(inst)
	h_move(inst)
	h_collide(inst)
	v_move_topdown(inst)
	v_collide_blocks(inst)
	go.set_position(inst.MY_POS, inst.MY_URL)

	--Reset buttons
	inst.buttons = 0
end

function physics_update_platformer(inst)
	get_xinput(inst)
	h_move(inst)
	h_collide(inst)
	ground_check_slopes(inst)
	v_move_platformer(inst)
	v_collide_slopes(inst)
	go.set_position(inst.MY_POS, inst.MY_URL)
	
	--Reset buttons
	inst.buttons = 0
end
